"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Processor = void 0;
const tslib_1 = require("tslib");
const t = tslib_1.__importStar(require("@babel/types"));
const core_1 = require("@taddy/core");
const taddy_1 = require("taddy");
const evaluate_1 = require("./helpers/evaluate");
const getObjectPropertyKey_1 = require("./helpers/getObjectPropertyKey");
const optimizeStaticStyles_1 = require("./helpers/optimizeStaticStyles");
const mergeObjectProperties_1 = require("./helpers/mergeObjectProperties");
const TSProcessor_1 = require("./TSProcessor");
function isUndefined(path) {
    return path.isIdentifier() && path.node.name === 'undefined';
}
function getLiteralValue(path) {
    try {
        return new Function(`return ${path.toString()}`)();
    }
    catch (error) {
        return getLiteralValue.FAIL;
    }
}
getLiteralValue.FAIL = Symbol('FAILED_VALUE');
function getHashedName(key, { postfix }) {
    return core_1.config.nameGenerator.getName(key, '', { postfix }).join('');
}
const isCSSProperty = (key) => {
    return key.slice(0, 2) === '--' || /[\w-]/.test(key);
};
// TODO: bottom up and split the initialization
let cachedTSProject;
class Processor {
    config;
    options;
    isStatic;
    classNameNodes;
    optimizationPaths;
    variables;
    prepare(options) {
        this.options = options;
        this.variables = [];
        this.classNameNodes = new WeakSet();
        this.optimizationPaths = new Set();
    }
    constructor({ config }) {
        this.config = config;
    }
    get tsProject() {
        if (!this.config.typescript) {
            throw new Error('TADDY: cant use typescript without enabled option');
        }
        return (cachedTSProject ||
            (cachedTSProject = new TSProcessor_1.TSProcessor(this.config.typescript === true ? {} : this.config.typescript)));
    }
    $css(rule, options) {
        // console.log('sm', this.options.sourceMap)
        const result = (0, taddy_1.$css)(rule, {
            ...options,
            // hash: stringHash(this.options.filename),
        });
        // console.trace({rule, options, result});
        return result;
    }
    evaluate(currentPath) {
        return (0, evaluate_1.evaluate)(currentPath, { exec: this.config.evaluate });
    }
    isClassNameNode(node) {
        return this.classNameNodes.has(node);
    }
    classNamesToNode(classNames) {
        const properties = [];
        for (const key in classNames) {
            const classValue = classNames[key];
            const propName = t.stringLiteral(key);
            let propValue;
            if (typeof classNames[key] === 'object') {
                propValue = t.objectExpression(this.classNamesToNode(classNames[key]));
            }
            else if (typeof classNames[key] === 'boolean') {
                propValue = t.booleanLiteral(classValue);
            }
            else {
                propValue = t.stringLiteral(String(classValue));
            }
            const node = t.objectProperty(propName, propValue);
            this.classNameNodes.add(node);
            properties.push(node);
        }
        return properties;
    }
    stylesToNode(key, value, { postfix = '', at } = {}) {
        const { className } = this.$css({ [key]: value }, { postfix, at });
        const properties = this.classNamesToNode(className);
        return properties;
    }
    processAtRule(key, path, { postfix, properties }) {
        const atRuleName = key;
        const valuePath = path.get('value');
        // if it's not an object expression, we can treat it as a regular object value
        if (!valuePath.isObjectExpression()) {
            return this.processPropertyValue(key, path, { postfix, properties });
        }
        for (const propPath of valuePath.get('properties')) {
            if (propPath.isObjectProperty()) {
                this.processObjectProperty(propPath, {
                    postfix,
                    at: { name: atRuleName },
                    properties,
                });
                continue;
            }
            if (propPath.isSpreadElement()) {
                this.processSpreadElement(propPath, {
                    postfix,
                    at: { name: atRuleName },
                    properties,
                });
                continue;
            }
            properties.push(propPath.node);
        }
    }
    processPropertyValue(key, path, { postfix, at, properties }) {
        const keyPath = path.get('key');
        const valuePath = path.get('value');
        const tryLiteralValue = () => {
            if (!(valuePath.isLiteral() || isUndefined(valuePath)))
                return false;
            const value = getLiteralValue(valuePath);
            if (value === getLiteralValue.FAIL) {
                return false;
            }
            properties.push(...this.stylesToNode(key, value, { postfix, at }));
            return true;
        };
        const tryObjectExpressionValue = () => {
            if (!valuePath.isObjectExpression())
                return false;
            // we shouldn't share "properties", as it should construct a new style object
            this.process(valuePath, {
                postfix: postfix + key,
                at,
            });
            // TODO: add comments
            let currentStyles = [];
            for (const nestedProp of valuePath.node.properties) {
                if (this.isClassNameNode(nestedProp)) {
                    if (currentStyles.length) {
                        properties.push(t.objectProperty(keyPath.node, t.objectExpression(currentStyles)));
                        currentStyles = [];
                    }
                    properties.push(nestedProp);
                }
                else {
                    currentStyles.push(nestedProp);
                }
            }
            if (currentStyles.length) {
                properties.push(t.objectProperty(keyPath.node, t.objectExpression(currentStyles)));
            }
            return true;
        };
        const tryComposesValue = () => {
            if (!(key === 'composes' && valuePath.isArrayExpression())) {
                return false;
            }
            if (postfix !== '') {
                throw new Error('TADDY: composes could be used only on top level');
            }
            for (const elemPath of valuePath.get('elements')) {
                this.process(elemPath, { at, postfix, properties });
            }
            return false;
        };
        const tryEvaluateValue = () => {
            const result = this.evaluate(valuePath);
            if ('error' in result)
                return false;
            properties.push(...this.stylesToNode(key, result.value, { postfix, at }));
            this.optimizationPaths.add(path);
            return true;
        };
        const STOP = Symbol('STOP');
        const getTSValue = (tsValuePath) => {
            const tsType = this.tsProject?.getTypeAtPos(this.options.filename, this.options.code, tsValuePath.node.start || 0, tsValuePath.node.end || 0);
            if (!tsType)
                return STOP;
            return (0, TSProcessor_1.parseValue)(tsType.getType());
        };
        const getComposesTSValue = (path) => {
            const elements = [];
            elements[TSProcessor_1.STATIC_KEY] = true;
            for (const elementPath of path.get('elements')) {
                if (!elementPath)
                    return;
                const value = getTSValue(elementPath);
                if (value === STOP)
                    return;
                elements[TSProcessor_1.STATIC_KEY] =
                    elements[TSProcessor_1.STATIC_KEY] && (0, TSProcessor_1.isStaticValue)(value);
                elements.push(value);
            }
            return elements;
        };
        const tryTypescriptValue = () => {
            if (!this.config.typescript)
                return false;
            let tsValue;
            if (key === 'composes' && valuePath.isArrayExpression()) {
                tsValue = getComposesTSValue(valuePath);
            }
            else {
                tsValue = getTSValue(valuePath);
            }
            if (tsValue === STOP)
                return false;
            if ((0, TSProcessor_1.isStaticValue)(tsValue)) {
                properties.push(...this.stylesToNode(key, tsValue, { postfix, at }));
                this.optimizationPaths.add(path);
                return true;
            }
            // TODO: improve composes handling
            if (key === 'composes')
                return false;
            const variants = [].concat(tsValue);
            let isCompilable = true;
            // pre-generate classes for variants
            variants.forEach((v) => {
                if (v instanceof TSProcessor_1.DynamicType || !(0, TSProcessor_1.isStaticValue)(v)) {
                    isCompilable = false;
                    return;
                }
                this.$css({ [key]: v }, { postfix, at });
            });
            if (!isCompilable)
                return false;
            const propKey = getHashedName(key, { postfix, at });
            const hashFunctionNode = this.options.addImport('h');
            properties.push(t.objectProperty(t.identifier(propKey), t.callExpression(hashFunctionNode, [
                valuePath.node,
            ])));
            return true;
        };
        const tryCSSVariableValue = () => {
            if (!this.config.CSSVariableFallback)
                return false;
            // think about dynamic nested for mixins
            // in case of nested dynamic values in mixin we can't fallback them as custom properties
            if (this.options.mixin ||
                key === 'composes' ||
                !isCSSProperty(key)) {
                return false;
            }
            const value = getHashedName(key, { postfix, at });
            const dynamicValue = `--${value}`;
            this.variables.push(t.objectProperty(t.stringLiteral(dynamicValue), valuePath.node));
            properties.push(...this.stylesToNode(key, `var(${dynamicValue})`, {
                postfix,
                at,
            }));
            return true;
        };
        if (tryLiteralValue() ||
            tryObjectExpressionValue() ||
            tryComposesValue() ||
            tryEvaluateValue() ||
            tryTypescriptValue() ||
            tryCSSVariableValue()) {
            return;
        }
        properties.push(path.node);
    }
    processObjectProperty(path, { postfix = '', at, properties }) {
        const keyPath = path.get('key');
        const valuePath = path.get('value');
        const key = (0, getObjectPropertyKey_1.getObjectPropertyKey)(keyPath);
        const tryEvaluateKey = () => {
            const result = this.evaluate(keyPath);
            if ('error' in result)
                return false;
            keyPath.replaceWith(t.stringLiteral(result.value));
            this.processPropertyValue(result.value, path, {
                postfix,
                at,
                properties,
            });
            return true;
        };
        if (key === 'className' || key === 'style' || key === null) {
            if (tryEvaluateKey()) {
                return;
            }
            properties.push(path.node);
            return;
        }
        if (key[0] === '@') {
            this.processAtRule(key, path, { postfix, properties });
            return;
        }
        if (at?.name && !at?.query) {
            this.process(valuePath, {
                postfix,
                properties,
                at: { name: at.name, query: key },
            });
            return;
        }
        this.processPropertyValue(key, path, { postfix, at, properties });
    }
    processSpreadElement(path, { postfix = '', at, properties }) {
        const tryEvaluateSpreadElement = () => {
            const result = this.evaluate(path.get('argument'));
            if ('error' in result)
                return false;
            const { className } = this.$css(result.value, { postfix, at });
            properties.push(...this.classNamesToNode(className));
            this.optimizationPaths.add(path);
            return true;
        };
        const tryTypescriptSpreadElement = () => {
            if (!this.config.typescript)
                return false;
            const { argument } = path.node;
            const tsType = this.tsProject?.getTypeAtPos(this.options.filename, this.options.code, argument.start || 0, argument.end || 0);
            if (!tsType)
                return false;
            const styles = (0, TSProcessor_1.parseValue)(tsType.getType());
            if (!(0, TSProcessor_1.isStaticValue)(styles))
                return false;
            const { className } = this.$css(styles, { postfix, at });
            properties.push(...this.classNamesToNode(className));
            this.optimizationPaths.add(path);
            return true;
        };
        if (tryEvaluateSpreadElement() || tryTypescriptSpreadElement()) {
            return;
        }
        // we shouldn't share "properties", as it should construct a new style object
        this.process(path.get('argument'), { postfix, at });
        properties.push(path.node);
    }
    processLogicalExpression(path, options) {
        this.process(path.get('left'), options);
        this.process(path.get('right'), options);
    }
    processObjectExpression(path, { postfix = '', at, properties = [] }) {
        for (const propPath of path.get('properties')) {
            if (propPath.isObjectProperty()) {
                this.processObjectProperty(propPath, { postfix, at, properties });
                continue;
            }
            if (propPath.isSpreadElement()) {
                this.processSpreadElement(propPath, { postfix, at, properties });
                continue;
            }
            properties.push(propPath.node);
        }
        /**
         * TODO: think about static optimizations for mixins
         */
        if (this.options.mixin)
            return;
        path.node.properties = (0, mergeObjectProperties_1.mergeObjectProperties)(properties);
    }
    processConditionalExpression(path, options = {}) {
        this.process(path.get('consequent'), options);
        this.process(path.get('alternate'), options);
    }
    process(path, options = {}) {
        if (path.isLogicalExpression()) {
            this.processLogicalExpression(path, options);
        }
        else if (path.isObjectExpression()) {
            this.processObjectExpression(path, options);
        }
        else if (path.isConditionalExpression()) {
            this.processConditionalExpression(path, options);
        }
    }
    run(callPath, options) {
        this.prepare(options);
        const tryEvaluate = (path) => {
            const result = this.evaluate(path);
            if ('error' in result)
                return false;
            const { className } = this.$css(result.value);
            /**
             * TODO: think about static optimizations for mixins
             */
            if (this.options.mixin)
                return true;
            const properties = this.classNamesToNode(className);
            path.replaceWith(t.objectExpression((0, mergeObjectProperties_1.mergeObjectProperties)(properties)));
            this.optimizationPaths.add(path);
            return true;
        };
        const args = [].concat(callPath.get('arguments'));
        let isObjectable = true;
        for (const arg of args) {
            if (tryEvaluate(arg)) {
                continue;
            }
            this.process(arg);
            if (!arg.isObjectExpression()) {
                isObjectable = false;
                continue;
            }
        }
        const nodes = args.map((x) => x.node);
        const argPath = args[0];
        if (this.variables.length > 0) {
            nodes.push(t.objectExpression([
                t.objectProperty(t.identifier(core_1.VARS_KEY), t.objectExpression(this.variables)),
            ]));
        }
        if (isObjectable) {
            argPath.replaceWith(t.objectExpression((0, mergeObjectProperties_1.mergeObjects)(nodes)));
            callPath.node.arguments = [argPath.node];
            const path = argPath;
            const isStatic = path.node.properties.every((x) => {
                if (!t.isObjectProperty(x))
                    return false;
                if (!('value' in x.key))
                    return false;
                if (!(x.key.value[0] === '_' ||
                    x.key.value === 'className' ||
                    x.key.value === 'style' ||
                    x.key.value === core_1.VARS_KEY ||
                    t.isBooleanLiteral(x.value))) {
                    return false;
                }
                return true;
            });
            if (!this.options.mixin && isStatic) {
                (0, optimizeStaticStyles_1.optimizeStaticStyles)(path);
            }
            return {
                isStatic,
                optimizationPaths: this.optimizationPaths,
            };
        }
        return {
            isStatic: false,
            optimizationPaths: this.optimizationPaths,
        };
    }
}
exports.Processor = Processor;
//# sourceMappingURL=Processor.js.map