"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.TSProcessor = exports.parseObject = exports.parseValue = exports.isStaticValue = exports.STATIC_KEY = exports.UnionType = exports.DynamicType = exports.getType = void 0;
const tslib_1 = require("tslib");
const path = tslib_1.__importStar(require("path"));
function getType(symbol) {
    return symbol?.getTypeAtLocation(symbol.getDeclarations()[0]);
}
exports.getType = getType;
class DynamicType {
    value;
    type;
    constructor({ value, type }) {
        this.value = value;
        this.type = type;
    }
}
exports.DynamicType = DynamicType;
class UnionType extends Array {
}
exports.UnionType = UnionType;
function isRecord(type) {
    return type?.getAliasSymbol()?.getEscapedName() === 'Record';
}
exports.STATIC_KEY = Symbol('__TADDY_STATIC_KEY__');
function isStaticValue(value) {
    if (!value || typeof value !== 'object') {
        return true;
    }
    if (value instanceof UnionType) {
        return false;
    }
    return value[exports.STATIC_KEY] || false;
}
exports.isStaticValue = isStaticValue;
function filterValue(value) {
    // filter vendor prefixes
    return String(value).startsWith('-');
}
function parseValue(type) {
    if (!type)
        return null;
    if (type.isAny() || type.isUnknown()) {
        return new DynamicType({ value: null, type });
    }
    if (isRecord(type)) {
        return parseRecord(type);
    }
    if (type.isArray()) {
        return parseArray(type);
    }
    if (type.isUnion()) {
        return new UnionType(...type.getUnionTypes().map(parseValue));
    }
    if (type.isObject()) {
        return parseObject(type);
    }
    let value = type?.getText();
    /**
     * Firstly, exclude types like "string & {}"
     *
     * TODO: think about other intersection types
     */
    if (!value || value?.includes('&')) {
        return null;
    }
    try {
        value = JSON.parse(value);
    }
    catch (e) {
        return new DynamicType({ value, type });
    }
    if (filterValue(value))
        return null;
    return value;
}
exports.parseValue = parseValue;
function parseMixinParameters(params) {
    const type = getType(params);
    return parseValue(type);
}
function parseMixin(mixin) {
    const type = getType(mixin);
    if (type?.getCallSignatures().length === 0) {
        return parseMixinParameters(mixin);
    }
    return type
        ?.getCallSignatures()[0]
        .getParameters()
        .map(parseMixinParameters);
}
function parseRecord(type) {
    const [k, v] = type.getAliasTypeArguments();
    const key = parseValue(k);
    const value = parseValue(v);
    if (!Array.isArray(key)) {
        return { [key]: value, [exports.STATIC_KEY]: value[exports.STATIC_KEY] };
    }
    return key.map((prop) => ({
        [prop]: value,
        [exports.STATIC_KEY]: value[exports.STATIC_KEY],
    }));
}
// TODO: improve the array parsing (such as not just union types)
function parseArray(type) {
    const elements = [];
    elements[exports.STATIC_KEY] = true;
    function applyElement(element) {
        const value = parseValue(element);
        if (!value)
            return;
        const isStatic = isStaticValue(value);
        elements[exports.STATIC_KEY] = elements[exports.STATIC_KEY] && isStatic;
        elements.push(isStatic ? value : new Set([].concat(value).filter(Boolean)));
    }
    const elementType = type.getArrayElementTypeOrThrow();
    for (const element of elementType.getUnionTypes() || [elementType]) {
        applyElement(element);
    }
    return elements;
}
function parseObject(type) {
    const props = {};
    let isStaticObject = true;
    for (const propSymbol of type?.getProperties() ?? []) {
        const name = propSymbol.getEscapedName();
        const value = parseMixin(propSymbol);
        if (!value)
            continue;
        const isStatic = isStaticValue(value);
        isStaticObject = isStaticObject && isStatic;
        props[name] = isStatic
            ? value
            : new Set([].concat(value).filter(Boolean));
    }
    props[exports.STATIC_KEY] = isStaticObject;
    return props;
}
exports.parseObject = parseObject;
class TSProcessor {
    project;
    constructor(config) {
        // lazy require for ts-morph
        const { Project } = require('ts-morph');
        if (config instanceof Project) {
            this.project = config;
            return;
        }
        this.project = new Project({
            tsConfigFilePath: path.join(process.cwd(), 'tsconfig.json'),
            ...config,
        });
    }
    getTypeAtPos(filename, code, start, end) {
        let sourceFile = this.project.getSourceFile(filename);
        if (sourceFile && sourceFile.getFullText() !== code) {
            sourceFile.replaceText([0, sourceFile.getEnd()], code);
        }
        else {
            sourceFile = this.project.createSourceFile(filename, code, {
                overwrite: true,
            });
        }
        const typeNode = sourceFile.getDescendantAtStartWithWidth(start, end - start);
        return typeNode;
    }
}
exports.TSProcessor = TSProcessor;
//# sourceMappingURL=TSProcessor.js.map