"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.findBindings = exports.addBinding = exports.BindingError = void 0;
const cache = new WeakMap();
class BindingError extends Error {
}
exports.BindingError = BindingError;
function isIterable(value) {
    return Symbol.iterator in Object(value);
}
function addBinding(bindingMap, binding, path) {
    if (isIterable(path)) {
        for (const x of path) {
            addBinding(bindingMap, binding, x);
        }
        return;
    }
    bindingMap.set(binding, (bindingMap.get(binding) || new Set()).add(path));
}
exports.addBinding = addBinding;
function isFunctionArgument(path) {
    return path.listKey === 'params' && path.parentPath?.isFunction();
}
function findBindings(currentPath, { throwError = false } = {}) {
    function visitIdentifier(referentPath, path, bindingMap) {
        if (path.parentPath.isMemberExpression()) {
            // @ts-expect-error TODO: fix
            path = path.parentPath.get('object');
        }
        else if (path.parentPath.isObjectPattern()) {
            // @ts-expect-error TODO: fix
            path = path.parentPath.parentPath;
        }
        const { parentPath, node } = path;
        if (!(parentPath.isExpression() ||
            parentPath.isSpreadElement() ||
            (parentPath.isVariableDeclarator() &&
                parentPath.node.init === node) ||
            (parentPath.isObjectProperty() &&
                (parentPath.node.value === node ||
                    parentPath.node.computed)))) {
            return;
        }
        if (!('name' in node && typeof node.name === 'string'))
            return;
        const { name } = node;
        const sharedBinding = currentPath.scope.getBinding(name);
        const binding = referentPath.scope.getBinding(name);
        if (!binding)
            return;
        if (sharedBinding !== binding)
            return;
        if (binding.constantViolations.length > 0) {
            throw new BindingError('CONSTANT VIOLATION');
        }
        const { path: bindingPath } = binding;
        if (bindingMap.has(binding)) {
            addBinding(bindingMap, binding, path);
            return;
        }
        if (isFunctionArgument(bindingPath)) {
            if (throwError) {
                throw new BindingError('FUNCTION ARGUMENT');
            }
            return;
        }
        addBinding(bindingMap, binding, path);
        // if (!(bindingPath.listKey === 'params')) {
        //     addBinding(bindingMap, binding, path);
        // }
        if (bindingPath.isImportSpecifier()) {
            return;
        }
        for (const [newBinding, paths] of traverse(bindingPath)) {
            addBinding(bindingMap, newBinding, paths);
        }
    }
    const visited = new Set();
    function traverse(referentPath) {
        if (referentPath.isObjectPattern()) {
            // throw new BindingError('OBJECT PATTERN');
        }
        if (cache.has(referentPath)) {
            return cache.get(referentPath);
        }
        const bindings = new Map();
        // we should not revisit the paths, otherwise something went wrong
        if (visited.has(referentPath)) {
            return bindings;
        }
        visited.add(referentPath);
        if (referentPath.isIdentifier()) {
            visitIdentifier(referentPath, referentPath, bindings);
        }
        referentPath.traverse({
            Identifier(path) {
                visitIdentifier(referentPath, path, bindings);
            },
        });
        cache.set(referentPath, bindings);
        return bindings;
    }
    return traverse(currentPath);
}
exports.findBindings = findBindings;
//# sourceMappingURL=findBindings.js.map